﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using FarPoint.Win.Spread;
using FarPoint.Win.Spread.CellType;

namespace DecimalNumber
{
    internal enum EColumns : int
    {
        HiddenInternalName = 0,
        HiddenNumber,
        DisplayName,
        Tripped,
        Reactivate,
        Activate,
        ActiveState,
        Apply,
        UpperLowerValue,
        Action,
        HiddenModeInternalName,
        ModeDisplayName,
        HiddenSensorInternalName,
        SensorDisplayName,
    };

    internal enum ECellTypes : int
    {
        ReadOnlyText,
        CheckBox,
        NumberAndUnit,
        ComboUpDown,
        ComboAction,
    }

    internal class ERMSheetView : FarPoint.Win.Spread.SheetView
    {
        public EDisplayType DisplayType { get; protected set; }

        protected LineDescription _LineDescription = new LineDescription();

        protected string Name { get; set; }

        protected int CountColumns { get; set; }
        protected int CountRows { get; set; }

        public int NumberOfDigits
        {
            set
            {
                foreach (ColumnDescription columnDescription in _LineDescription.Values)
                {
                    if (columnDescription.ECellType == ECellTypes.NumberAndUnit)
                    {
                        for (int i = 0; i < RowCount; i++)
                        {
                            if (Cells.Get(i, columnDescription.Column).CellType is NumberUnitCellType)
                            {
                                (Cells.Get(i, columnDescription.Column).CellType as NumberUnitCellType).NumberOfDigits = value;
                            }
                            else
                            {
                                System.Diagnostics.Debug.Assert(false);
                            }
                        }
                    }
                }
            }
        }

        internal ERMSheetView()
        {
            ReferenceStyle = FarPoint.Win.Spread.Model.ReferenceStyle.R1C1;

            Name = "Hallo";
            CountColumns = 10;
            CountRows = 0;

            DisplayType = EDisplayType.Summary;
        }

        internal void Initialize()
        {
            Reset();
            SheetName = Name;
            ColumnCount = _LineDescription.ColumnCount;
            RowCount = 0;

            //SelectionUnit = FarPoint.Win.Spread.Model.SelectionUnit.Row;
            //SelectionPolicy = FarPoint.Win.Spread.Model.SelectionPolicy.MultiRange;
            //OperationMode = FarPoint.Win.Spread.OperationMode.ReadOnly;

            SetDesign();
        }

        private void SetDesign()
        {
            foreach (EColumns column in _LineDescription.Keys)
            {
                ColumnDescription columnDescription = _LineDescription[column];
                Columns[columnDescription.Column].Visible = columnDescription.Visibility;
                
                if (columnDescription.HasHeaderToDisplay)
                {
                    Models.ColumnHeaderData.SetValue(0, columnDescription.Column, TBD_Translation(columnDescription.HeaderText));
                }
            }
        }

        private string TBD_Translation(string toTranslate)
        {
            return toTranslate;
        }

        internal virtual void AddModelItem(ModelItem item) { }

        internal virtual void RemoveModelItem(ModelItem item) { }

        protected void RemoveMainEntry(ModelItem item)
        {
            int rowMainEntry = FindRowMainEntry(item.MainEntry);

            if (rowMainEntry != -1)
            {
                RemoveRows(rowMainEntry, GetCurrentRowSpan(rowMainEntry));
            }
        }

        protected int FindRowMainEntry(string mainEntry)
        {
            int column = _LineDescription[EColumns.HiddenInternalName].Column;

            for (int row = 0; row < RowCount; row++)
            {
                if (Cells.Get(row, column).Text == mainEntry)
                {
                    return row;
                }
            }

            return -1;
        }

        protected int GetCurrentRowSpan(int rowMainEntry)
        {
            return RowHeader.Cells[rowMainEntry, 0, rowMainEntry, 0].RowSpan;
        }
    }

    internal class SheetViewMainSub : ERMSheetView
    {
        internal SheetViewMainSub() { }

        internal override void AddModelItem(ModelItem item)
        {
            int rowSubEntry = -1;

            int rowMainEntry = FindRowMainEntry(item.MainEntry);

            if (rowMainEntry < 0)
            {
                rowMainEntry = RowCount;
                rowSubEntry = RowCount;

                AddRows(RowCount, 1);
                Rows[rowMainEntry].Resizable = false;
                SetCellTypes(item, rowMainEntry, rowSubEntry);
            }
            else
            {
                rowSubEntry = FindRowSubEntry(item.SubEntry, rowMainEntry);
                if (-1 == rowSubEntry)
                {
                    rowSubEntry = AppendSubEntry(rowMainEntry);
                    SetCellTypes(item, rowMainEntry, rowSubEntry);
                }
            }

            ModifyData(item, rowMainEntry, rowSubEntry);
        }

        internal override void RemoveModelItem(ModelItem item)
        {
            int rowMainEntry = FindRowMainEntry(item.MainEntry);

            if (rowMainEntry < 0)
            {
                System.Diagnostics.Debug.Assert(false);
                return;
            }

            if (1 < GetCurrentRowSpan(rowMainEntry))
            {
                RemoveSubEntry(rowMainEntry, item);
            }
            else
            {
                RemoveMainEntry(item);
            }
        }
        
        protected virtual void ModifyData(ModelItem item, int rowMainEntry, int rowSubEntry) { }

        private int AppendSubEntry(int rowMainEntry)
        {
            int rowSpan = GetCurrentRowSpan(rowMainEntry);

            AddRows(rowMainEntry + rowSpan, 1);
            Rows[rowMainEntry + rowSpan].Resizable = false;
            AddRowHeaderSpanCell(rowMainEntry, 0, rowSpan + 1, 1);

            foreach (ColumnDescription columnDescription in _LineDescription.Values)
            {
                if (columnDescription.RowsSpanned)
                {
                    Cell mainCell = Cells.Get(rowMainEntry, columnDescription.Column);
                    mainCell.RowSpan = rowSpan + 1;
                }
            }

            return rowMainEntry + rowSpan;
        }

        private void RemoveSubEntry(int rowMainEntry, ModelItem item)
        {
            int currentRowSpan = GetCurrentRowSpan(rowMainEntry);

            System.Diagnostics.Debug.Assert(currentRowSpan > 1, "Remove the MainEntry instead!");

            int rowSubEntry = FindRowSubEntry(item.SubEntry, rowMainEntry);

            Dictionary<int, object> MainRowValues = new Dictionary<int, object>();

            foreach (ColumnDescription columnDescription in _LineDescription.Values)
            {
                if (columnDescription.RowsSpanned)
                {
                    MainRowValues.Add(columnDescription.Column, Cells.Get(rowMainEntry, columnDescription.Column).Value);
                }
            }

            RemoveRows(rowSubEntry, 1);
            RemoveRowHeaderSpanCell(rowMainEntry, 1);

            foreach (int column in MainRowValues.Keys)
            {
                Cell mainCell = Cells.Get(rowMainEntry, column);
                mainCell.RowSpan = currentRowSpan - 1;
                mainCell.Value = MainRowValues[column];
            }
        }

        private int FindRowSubEntry(string subEntry, int rowMainEntry)
        {
            int column = _LineDescription[EColumns.HiddenSensorInternalName].Column;

            int currentRowSpan = GetCurrentRowSpan(rowMainEntry);

            for (int row = rowMainEntry; row < rowMainEntry + currentRowSpan; row++)
            {
                if (Cells.Get(row, column).Text == subEntry)
                {
                    return row;
                }
            }

            return -1;
        }

        private void SetCellTypes(ModelItem item, int rowMainEntry, int rowSubEntry)
        {
            foreach (EColumns eColumn in _LineDescription.Keys)
            {
                ColumnDescription columnDescription = _LineDescription[eColumn];

                if (columnDescription.RowsSpanned)
                {
                    Cell topCell = Cells.Get(rowMainEntry, columnDescription.Column);
                    item.SetCellType(ref topCell, columnDescription.ECellType);
                }
                else
                {
                    Cell subCell = Cells.Get(rowSubEntry, columnDescription.Column);
                    item.SetCellType(ref subCell, columnDescription.ECellType);
                }
            }
        }
    }

    internal class SheetViewRateOfChange : SheetViewMainSub
    {
        internal SheetViewRateOfChange()
        {
            _LineDescription.Clear();
            _LineDescription.CountRows = 1;

            int col = 0;
            _LineDescription.Add(EColumns.HiddenInternalName, new ColumnDescription(col++, false, "HiddenInternalName", ECellTypes.ReadOnlyText, true));         // controllerInternalName
            _LineDescription.Add(EColumns.HiddenNumber, new ColumnDescription(col++, false, "HiddenNumber", ECellTypes.ReadOnlyText, true));
            _LineDescription.Add(EColumns.DisplayName, new ColumnDescription(col++, true, "Controller", ECellTypes.ReadOnlyText, true));                         // controllerDisplayName
            _LineDescription.Add(EColumns.HiddenModeInternalName, new ColumnDescription(col++, false, "HiddenModeInternalName", ECellTypes.ReadOnlyText));       // internalName
            _LineDescription.Add(EColumns.ModeDisplayName, new ColumnDescription(col++, true, "Mode", ECellTypes.ReadOnlyText));                                 // displayName
            _LineDescription.Add(EColumns.HiddenSensorInternalName, new ColumnDescription(col++, false, "HiddenSensorInternalName", ECellTypes.ReadOnlyText));   // sensorInternalName
            _LineDescription.Add(EColumns.SensorDisplayName, new ColumnDescription(col++, true, "Sensor", ECellTypes.ReadOnlyText));                             // sensorDisplayName
            _LineDescription.Add(EColumns.Tripped, new ColumnDescription(col++, true, "Tripped", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.Activate, new ColumnDescription(col++, true, "Activate", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.UpperLowerValue, new ColumnDescription(col++, true, "Value", ECellTypes.NumberAndUnit));
            _LineDescription.Add(EColumns.Action, new ColumnDescription(col++, true, "Action", ECellTypes.ComboAction));

            Name = "Änderungsüberwachung";

            DisplayType = EDisplayType.RateOfChange;
        }

        protected override void ModifyData(ModelItem item, int rowMainEntry, int rowSubEntry)
        {
            if (false == (item is ModelItemRateOfChange))
            {
                System.Diagnostics.Debug.Assert(false);
                return;
            }

            ModelItemRateOfChange rocItem = item as ModelItemRateOfChange;

            // Controller
            Cells.Get(rowMainEntry, _LineDescription[EColumns.HiddenInternalName].Column).Value = rocItem.MainEntry; // = ControllerInternalName
            Cells.Get(rowMainEntry, _LineDescription[EColumns.DisplayName].Column).Value = rocItem.ControllerDisplayName;

            // Mode
            Cells.Get(rowSubEntry, _LineDescription[EColumns.HiddenModeInternalName].Column).Value = rocItem.InternalName;   // Mode
            Cells.Get(rowSubEntry, _LineDescription[EColumns.ModeDisplayName].Column).Value = rocItem.DisplayName;

            // Sensor
            Cells.Get(rowSubEntry, _LineDescription[EColumns.HiddenSensorInternalName].Column).Value = rocItem.SubEntry; // SensorInternalName
            Cells.Get(rowSubEntry, _LineDescription[EColumns.SensorDisplayName].Column).Value = rocItem.SensorDisplayName;

            Cells.Get(rowSubEntry, _LineDescription[EColumns.Activate].Column).Value = rocItem.Arming;

            Cells.Get(rowSubEntry, _LineDescription[EColumns.UpperLowerValue].Column).Value = rocItem.Percentage;

            Cells.Get(rowSubEntry, _LineDescription[EColumns.Tripped].Column).Value = rocItem.Tripped;
            Cells.Get(rowSubEntry, _LineDescription[EColumns.Action].Column).Value = new ComboContentERMActions(rocItem.EAction);
        }
    }

    internal class SheetViewFixedSize : ERMSheetView
    {

        internal override void AddModelItem(ModelItem item)
        {
            int rowEntry = FindRowMainEntry(item.MainEntry);

            if (rowEntry < 0)
            {
                rowEntry = RowCount;

                AddRows(rowEntry, _LineDescription.CountRows);
                for (int i = 0; i < _LineDescription.CountRows; i++)
                {
                    Rows[rowEntry + i].Resizable = false;
                }
                AddRowHeaderSpanCell(rowEntry, 0, _LineDescription.CountRows, 1);
                //AppendEntry(item);
                SetCellTypes(item, rowEntry);
            }

            ModifyData(item, rowEntry);
        }

        private void SetCellTypes(ModelItem item, int rowMainEntry)
        {
            foreach (EColumns eColumn in _LineDescription.Keys)
            {
                ColumnDescription columnDescription = _LineDescription[eColumn];

                if (columnDescription.RowsSpanned)
                {
                    Cell topCell = Cells.Get(rowMainEntry, columnDescription.Column);
                    topCell.RowSpan = _LineDescription.CountRows;
                    item.SetCellType(ref topCell, columnDescription.ECellType);
                }
                else
                {
                    for (int row = 0; row < _LineDescription.CountRows; row++)
                    {
                        Cell subCell = Cells.Get(rowMainEntry + row, columnDescription.Column);
                        item.SetCellType(ref subCell, columnDescription.ECellType);
                    }
                }
            }
        }

        internal override void RemoveModelItem(ModelItem item)
        {
            int rowMainEntry = FindRowMainEntry(item.MainEntry);

            if (rowMainEntry < 0)
            {
                System.Diagnostics.Debug.Assert(false);
                return;
            }

            RemoveMainEntry(item);
        }

        protected virtual void ModifyData(ModelItem item, int rowMainEntry) { }
    }

    internal class SheetViewLimits : SheetViewFixedSize
    {
        internal SheetViewLimits()
        {
            Name = "Grenzwerte";

            DisplayType = EDisplayType.Limits;

            _LineDescription.Clear();
            _LineDescription.CountRows = 2;

            int col = 0;
            _LineDescription.Add(EColumns.HiddenInternalName, new ColumnDescription(col++, false, "HiddenInternalName", ECellTypes.ReadOnlyText, true));
            _LineDescription.Add(EColumns.HiddenNumber, new ColumnDescription(col++, false, "HiddenNumber", ECellTypes.ReadOnlyText, true));
            _LineDescription.Add(EColumns.DisplayName, new ColumnDescription(col++, true, "Source", ECellTypes.ReadOnlyText, true));
            _LineDescription.Add(EColumns.Tripped, new ColumnDescription(col++, true, "Tripped", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.Reactivate, new ColumnDescription(col++, true, "Reactivate", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.Activate, new ColumnDescription(col++, true, "Activate", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.UpperLowerValue, new ColumnDescription(col++, true, "Value", ECellTypes.NumberAndUnit));
            _LineDescription.Add(EColumns.Action, new ColumnDescription(col++, true, "Action", ECellTypes.ComboAction));
        }

        protected override void ModifyData(ModelItem item, int rowMainEntry)
        {
            if (false == (item is ModelItemLimit))
            {
                System.Diagnostics.Debug.Assert(false);
                return;
            }

            ModelItemLimit itemLimit = item as ModelItemLimit;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.HiddenInternalName].Column).Value = itemLimit.InternalName;    // = InternalName
            Cells.Get(rowMainEntry, _LineDescription[EColumns.DisplayName].Column).Value = itemLimit.DisplayName;            // = InternalName

            Cells.Get(rowMainEntry, _LineDescription[EColumns.Activate].Column).Value = itemLimit.UpperArming;
            Cells.Get(rowMainEntry + 1, _LineDescription[EColumns.Activate].Column).Value = itemLimit.LowerArming;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.Reactivate].Column).Value = itemLimit.UpperRearming;
            Cells.Get(rowMainEntry + 1, _LineDescription[EColumns.Reactivate].Column).Value = itemLimit.LowerRearming;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.Tripped].Column).Value = itemLimit.UpperTripped;
            Cells.Get(rowMainEntry + 1, _LineDescription[EColumns.Tripped].Column).Value = itemLimit.LowerTripped;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.UpperLowerValue].Column).Value = itemLimit.UpperValue;
            Cells.Get(rowMainEntry + 1, _LineDescription[EColumns.UpperLowerValue].Column).Value = itemLimit.LowerValue;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.Action].Column).Value = new ComboContentERMActions(itemLimit.UpperEAction);
            Cells.Get(rowMainEntry + 1, _LineDescription[EColumns.Action].Column).Value = new ComboContentERMActions(itemLimit.LowerEAction);
        }
    }

    internal class SheetViewThresholds : SheetViewFixedSize
    {
        internal SheetViewThresholds()
        {
            Name = "Schwellwerte";

            DisplayType = EDisplayType.Thresholds;

            _LineDescription.Clear();
            _LineDescription.CountRows = 2;

            int col = 0;
            _LineDescription.Add(EColumns.HiddenInternalName, new ColumnDescription(col++, false, "HiddenInternalName", ECellTypes.ReadOnlyText, true));
            _LineDescription.Add(EColumns.HiddenNumber, new ColumnDescription(col++, false, "HiddenNumber", ECellTypes.ReadOnlyText, true));
            _LineDescription.Add(EColumns.DisplayName, new ColumnDescription(col++, true, "Source", ECellTypes.ReadOnlyText, true));
            _LineDescription.Add(EColumns.Tripped, new ColumnDescription(col++, true, "Tripped", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.Reactivate, new ColumnDescription(col++, true, "Reactivate", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.Activate, new ColumnDescription(col++, true, "Activate", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.Apply, new ColumnDescription(col++, true, "Apply", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.UpperLowerValue, new ColumnDescription(col++, true, "Value", ECellTypes.NumberAndUnit));
            _LineDescription.Add(EColumns.Action, new ColumnDescription(col++, true, "Action", ECellTypes.ComboAction));
        }

        protected override void ModifyData(ModelItem item, int rowMainEntry)
        {
            if (false == (item is ModelItemThreshold))
            {
                System.Diagnostics.Debug.Assert(false);
                return;
            }

            ModelItemThreshold itemThreshold = item as ModelItemThreshold;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.HiddenInternalName].Column).Value = itemThreshold.InternalName;    // = InternalName
            Cells.Get(rowMainEntry, _LineDescription[EColumns.DisplayName].Column).Value = itemThreshold.DisplayName;            // = InternalName

            Cells.Get(rowMainEntry, _LineDescription[EColumns.Activate].Column).Value = itemThreshold.UpperArming;
            Cells.Get(rowMainEntry + 1, _LineDescription[EColumns.Activate].Column).Value = itemThreshold.LowerArming;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.Reactivate].Column).Value = itemThreshold.UpperRearming;
            Cells.Get(rowMainEntry + 1, _LineDescription[EColumns.Reactivate].Column).Value = itemThreshold.LowerRearming;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.Apply].Column).Value = itemThreshold.UseUpperValue;
            Cells.Get(rowMainEntry + 1, _LineDescription[EColumns.Apply].Column).Value = itemThreshold.UseLowerValue;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.Tripped].Column).Value = itemThreshold.UpperTripped;
            Cells.Get(rowMainEntry + 1, _LineDescription[EColumns.Tripped].Column).Value = itemThreshold.LowerTripped;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.UpperLowerValue].Column).Value = itemThreshold.UpperValue;
            Cells.Get(rowMainEntry + 1, _LineDescription[EColumns.UpperLowerValue].Column).Value = itemThreshold.LowerValue;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.Action].Column).Value = new ComboContentERMActions(itemThreshold.UpperEAction);
            Cells.Get(rowMainEntry + 1, _LineDescription[EColumns.Action].Column).Value = new ComboContentERMActions(itemThreshold.LowerEAction);
        }

    }

    internal class SheetViewDigitalInputs : SheetViewFixedSize
    {
        internal SheetViewDigitalInputs()
        {
            Name = "DigitaleEingaenge";

            DisplayType = EDisplayType.DigitalInputs;

            _LineDescription.Clear();
            _LineDescription.CountRows = 1;

            int col = 0;
            _LineDescription.Add(EColumns.HiddenInternalName, new ColumnDescription(col++, false, "HiddenInternalName", ECellTypes.ReadOnlyText, true));
            _LineDescription.Add(EColumns.HiddenNumber, new ColumnDescription(col++, false, "HiddenNumber", ECellTypes.ReadOnlyText, true));
            _LineDescription.Add(EColumns.DisplayName, new ColumnDescription(col++, true, "Source", ECellTypes.ReadOnlyText, true));
            _LineDescription.Add(EColumns.Tripped, new ColumnDescription(col++, true, "Tripped", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.Activate, new ColumnDescription(col++, true, "Activate", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.ActiveState, new ColumnDescription(col++, true, "State", ECellTypes.ComboUpDown));
            _LineDescription.Add(EColumns.Action, new ColumnDescription(col++, true, "Action", ECellTypes.ComboAction));
        }

        protected override void ModifyData(ModelItem item, int rowMainEntry)
        {
            if (false == (item is ModelItemDigitalInput))
            {
                System.Diagnostics.Debug.Assert(false);
                return;
            }

            ModelItemDigitalInput itemDigitalInput = item as ModelItemDigitalInput;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.HiddenInternalName].Column).Value = itemDigitalInput.InternalName;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.DisplayName].Column).Value = itemDigitalInput.DisplayName;
            Cells.Get(rowMainEntry, _LineDescription[EColumns.Activate].Column).Value = itemDigitalInput.Arming;
            Cells.Get(rowMainEntry, _LineDescription[EColumns.ActiveState].Column).Value = new ComboContentEnumUpDown(itemDigitalInput.UpDown);
            Cells.Get(rowMainEntry, _LineDescription[EColumns.Tripped].Column).Value = itemDigitalInput.Tripped;
            Cells.Get(rowMainEntry, _LineDescription[EColumns.Action].Column).Value = new ComboContentERMActions(itemDigitalInput.EAction);
        }
    }

    internal class SheetViewLoopError : SheetViewFixedSize
    {
        internal SheetViewLoopError()
        {
            _LineDescription.Clear();
            _LineDescription.CountRows = 1;

            int col = 0;
            _LineDescription.Add(EColumns.HiddenInternalName, new ColumnDescription(col++, false, "HiddenInternalName", ECellTypes.ReadOnlyText, true));
            _LineDescription.Add(EColumns.HiddenNumber, new ColumnDescription(col++, false, "HiddenNumber", ECellTypes.ReadOnlyText, true));
            _LineDescription.Add(EColumns.DisplayName, new ColumnDescription(col++, true, "Source", ECellTypes.ReadOnlyText, true));
            _LineDescription.Add(EColumns.Tripped, new ColumnDescription(col++, true, "Tripped", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.Reactivate, new ColumnDescription(col++, true, "Reactivate", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.Activate, new ColumnDescription(col++, true, "Activate", ECellTypes.CheckBox));
            _LineDescription.Add(EColumns.UpperLowerValue, new ColumnDescription(col++, true, "Value", ECellTypes.NumberAndUnit));
            _LineDescription.Add(EColumns.Action, new ColumnDescription(col++, true, "Action", ECellTypes.ComboAction));

            Name = "Regelabweichung";

            DisplayType = EDisplayType.LoopErrors;
        }

        // when derived from MainSub
        //protected override void ModifyData(ModelItem item, int rowMainEntry, int rowSubEntry) { }

        protected override void ModifyData(ModelItem item, int rowMainEntry)
        {
            if (false == (item is ModelItemLoopError))
            {
                System.Diagnostics.Debug.Assert(false);
                return;
            }

            ModelItemLoopError itemLoopError = item as ModelItemLoopError;

            Cells.Get(rowMainEntry, _LineDescription[EColumns.HiddenInternalName].Column).Value = itemLoopError.InternalName;
            Cells.Get(rowMainEntry, _LineDescription[EColumns.DisplayName].Column).Value = itemLoopError.DisplayName;
            Cells.Get(rowMainEntry, _LineDescription[EColumns.Activate].Column).Value = itemLoopError.Arming;
            Cells.Get(rowMainEntry, _LineDescription[EColumns.Reactivate].Column).Value = itemLoopError.Rearming;
            Cells.Get(rowMainEntry, _LineDescription[EColumns.Tripped].Column).Value = itemLoopError.Tripped;
            Cells.Get(rowMainEntry, _LineDescription[EColumns.UpperLowerValue].Column).Value = itemLoopError.LoopErrorValue;
            Cells.Get(rowMainEntry, _LineDescription[EColumns.Action].Column).Value = new ComboContentERMActions(itemLoopError.EAction);
        }
    }

    internal class ColumnDescription
    {
        internal ECellTypes ECellType { get; private set; }

        internal string HeaderText { get; private set; }
        internal bool Visibility { get; private set; }

        internal int Column { get; private set; }
        internal bool RowsSpanned { get; private set; }

        internal bool HasHeaderToDisplay
        {
            get
            {
                return ((Visibility) && (HeaderText != null) && (HeaderText != string.Empty));
            }
        }

        internal ColumnDescription(int column, bool visibility, string headerText, ECellTypes eCellType, bool rowsSpanned = false)
        {
            Column = column;
            Visibility = visibility;
            HeaderText = headerText;
            ECellType = eCellType;
            RowsSpanned = rowsSpanned;
        }
    }

    internal class LineDescription : Dictionary<EColumns, ColumnDescription>
    {
        internal LineDescription() { }

        /// <summary>
        /// CountRows = 0 means variable count of rows, the count of rows has to be determined from the sheet.
        /// E.g. for "rate of change" where RowCount depends on the count of sensors connected to a controller.
        /// </summary>
        internal int CountRows { get; set; }

        internal int ColumnCount
        {
            get
            {
                int columnCount = 0;
                foreach (ColumnDescription columnDescription in Values)
                {
                    columnCount = Math.Max(columnCount, columnDescription.Column + 1);
                }

                return columnCount;
            }
        }
    }
}