﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using FarPoint.Win.Spread;
using FarPoint.Win.Spread.CellType;

namespace DecimalNumber
{
    internal enum EColumns : int
    {
        HiddenInternalName = 0,
        DisplayName,
        SomeValue,
    };

    internal enum ECellTypes : int
    {
        ReadOnlyText,
        CheckBox,
        NumberAndUnit,
        ComboUpDown,
        ComboAction,
    }

    internal class MySheetView : FarPoint.Win.Spread.SheetView
    {
        protected LineDescription _LineDescription = new LineDescription();

        protected string Name { get; set; }

        public int NumberOfDigits
        {
            set
            {
                foreach (ColumnDescription columnDescription in _LineDescription.Values)
                {
                    if (columnDescription.ECellType == ECellTypes.NumberAndUnit)
                    {
                        for (int i = 0; i < RowCount; i++)
                        {
                            if (Cells.Get(i, columnDescription.Column).CellType is NumberUnitCellType)
                            {
                                (Cells.Get(i, columnDescription.Column).CellType as NumberUnitCellType).NumberOfDigits = value;
                            }
                            else
                            {
                                System.Diagnostics.Debug.Assert(false);
                            }
                        }
                    }
                }
            }
        }

        internal MySheetView()
        {
            ReferenceStyle = FarPoint.Win.Spread.Model.ReferenceStyle.R1C1;

            Name = "Hello";
        }

        internal void Initialize()
        {
            Reset();
            SheetName = Name;
            ColumnCount = _LineDescription.CountColumn;
            RowCount = 0;

            SetDesign();
        }

        private void SetDesign()
        {
            foreach (EColumns column in _LineDescription.Keys)
            {
                ColumnDescription columnDescription = _LineDescription[column];
                Columns[columnDescription.Column].Visible = columnDescription.Visibility;
                
                if (columnDescription.HasHeaderToDisplay)
                {
                    Models.ColumnHeaderData.SetValue(0, columnDescription.Column, columnDescription.HeaderText);
                }
            }
        }

        internal virtual void AddModelItem(ModelItem item) { }
    }

    internal class SheetViewFixedSize : MySheetView
    {
        internal override void AddModelItem(ModelItem item)
        {
            int rowEntry = -1;

            if (rowEntry < 0)
            {
                rowEntry = RowCount;

                AddRows(rowEntry, _LineDescription.CountRows);
                for (int i = 0; i < _LineDescription.CountRows; i++)
                {
                    Rows[rowEntry + i].Resizable = false;
                }
                AddRowHeaderSpanCell(rowEntry, 0, _LineDescription.CountRows, 1);
                SetCellTypes(item, rowEntry);
            }

            ModifyData(item, rowEntry);
        }

        private void SetCellTypes(ModelItem item, int rowMainEntry)
        {
            foreach (EColumns eColumn in _LineDescription.Keys)
            {
                ColumnDescription columnDescription = _LineDescription[eColumn];

                if (columnDescription.RowsSpanned)
                {
                    Cell topCell = Cells.Get(rowMainEntry, columnDescription.Column);
                    topCell.RowSpan = _LineDescription.CountRows;
                    item.SetCellType(ref topCell, columnDescription.ECellType);
                }
                else
                {
                    for (int row = 0; row < _LineDescription.CountRows; row++)
                    {
                        Cell subCell = Cells.Get(rowMainEntry + row, columnDescription.Column);
                        item.SetCellType(ref subCell, columnDescription.ECellType);
                    }
                }
            }
        }

        protected virtual void ModifyData(ModelItem item, int rowMainEntry) { }
    }

    internal class SheetViewSpecial : SheetViewFixedSize
    {
        internal SheetViewSpecial()
        {
            _LineDescription.Clear();
            _LineDescription.CountRows = 1;

            int col = 0;
            _LineDescription.Add(EColumns.HiddenInternalName, new ColumnDescription(col++, false, "HiddenInternalName", ECellTypes.ReadOnlyText));
            _LineDescription.Add(EColumns.DisplayName, new ColumnDescription(col++, true, "Source", ECellTypes.ReadOnlyText));
            _LineDescription.Add(EColumns.SomeValue, new ColumnDescription(col++, true, "Value", ECellTypes.NumberAndUnit));
        }

        protected override void ModifyData(ModelItem item, int rowMainEntry)
        {
            Cells.Get(rowMainEntry, _LineDescription[EColumns.HiddenInternalName].Column).Value = item.InternalName;
            Cells.Get(rowMainEntry, _LineDescription[EColumns.DisplayName].Column).Value = item.DisplayName;
            Cells.Get(rowMainEntry, _LineDescription[EColumns.SomeValue].Column).Value = item.SomeValue;
        }
    }

    internal class ColumnDescription
    {
        internal ECellTypes ECellType { get; private set; }

        internal string HeaderText { get; private set; }
        internal bool Visibility { get; private set; }

        internal int Column { get; private set; }
        internal bool RowsSpanned { get; private set; }

        internal bool HasHeaderToDisplay
        {
            get
            {
                return ((Visibility) && (HeaderText != null) && (HeaderText != string.Empty));
            }
        }

        internal ColumnDescription(int column, bool visibility, string headerText, ECellTypes eCellType, bool rowsSpanned = false)
        {
            Column = column;
            Visibility = visibility;
            HeaderText = headerText;
            ECellType = eCellType;
            RowsSpanned = rowsSpanned;
        }
    }

    internal class LineDescription : Dictionary<EColumns, ColumnDescription>
    {
        internal LineDescription() { }

        internal int CountRows { get; set; }

        internal int CountColumn
        {
            get
            {
                int columnCount = 0;
                foreach (ColumnDescription columnDescription in Values)
                {
                    columnCount = Math.Max(columnCount, columnDescription.Column + 1);
                }

                return columnCount;
            }
        }
    }
}