﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DecimalNumber
{
    public class NumberUnitCellType : FarPoint.Win.Spread.CellType.PercentCellType
    {
        private static int ms_NumberOfDigits = 2;

        private static Dictionary<int, float> ms_DecimalIncrement = new Dictionary<int, float>()
        {
            {1, 0.1F},
            {2, 0.01F},
            {3, 0.001F},
            {4, 0.0001F},
            {5, 0.00001F},
        };

        public int NumberOfDigits
        {
            set
            {
                if (ms_DecimalIncrement.ContainsKey(value))
                {
                    ms_NumberOfDigits = value;
                    DecimalPlaces = ms_NumberOfDigits;
                    SpinDecimalIncrement = ms_DecimalIncrement[ms_NumberOfDigits];
                }
                else
                {
                    System.Diagnostics.Debug.Assert(false);
                }
            }
        }

        public string Unit
        {
            get
            {
                if (PercentSign.Length > 0)
                {
                    return PercentSign.Substring(1);
                }
                else
                {
                    return string.Empty;
                }
            }
            set
            {
                if ((value != null) && (value != string.Empty))
                {
                    PercentSign = " " + value;
                }
            }
        }

        public NumberUnitCellType()
        {
            SetDefaults();
        }

        public NumberUnitCellType(string unit)
        {
            SetDefaults();
            Unit = unit;
        }

        public NumberUnitCellType(NumberUnitCellType org)
        {
            SetDefaults();

            MinimumValue = org.MinimumValue;
            MaximumValue = org.MaximumValue;
            Unit = org.Unit;
        }

        private void SetDefaults()
        {
            EnableSubEditor = false;
            FixedPoint = true;
            SpinButton = true;
            ShowSeparator = false;
            NumberOfDigits = ms_NumberOfDigits;
        }

        public override object Clone()
        {
            return new NumberUnitCellType(this);
        }

        public void AdjustBoundary(bool upper, double boundary)
        {
            if (upper)
            {
                MaximumValue = boundary;
            }
            else
            {
                MinimumValue = boundary;
            }
        }
    }
}
